/*
 * Copyright (c) 2005 Landon J. Fuller <landonf@opendarwin.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright owner nor the names of contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>
#include <stdlib.h>

#include <fcntl.h>

#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>

#include <net/if.h>
#include <net/pfvar.h>

#include <errno.h>
#include <err.h>

const char pf_device[] = "/dev/pf";

enum { IN, OUT };
enum { IPV4, IPV6 };
enum { PASS, BLOCK };

void *xmalloc(size_t size) {
	void *ptr = malloc(size);
	if (!ptr) {
		err(1, "malloc failed!");
	}
	return (ptr);
}

void *xrealloc(void *ptr, size_t size) {
	ptr = realloc(ptr, size);
	if (!ptr) {
		err(1, "realloc failed!");
	}
	return (ptr);
}

void print_iface_stats(struct pfi_if *iface) {
	u_int64_t ipv4_in_pass, ipv4_in_block;
	u_int64_t ipv6_in_pass, ipv6_in_block;
	u_int64_t ipv4_out_pass, ipv4_out_block;
	u_int64_t ipv6_out_pass, ipv6_out_block;
	u_int64_t total_in_pass, total_in_block;
	u_int64_t total_out_pass, total_out_block;
	u_int64_t total_in, total_out;
	
	ipv4_in_pass = iface->pfif_bytes[IPV4][IN][PASS];
	ipv6_in_pass = iface->pfif_bytes[IPV6][IN][PASS];
	total_in_pass = ipv4_in_pass + ipv6_in_pass;

	ipv4_in_block = iface->pfif_bytes[IPV4][IN][BLOCK];
	ipv6_in_block = iface->pfif_bytes[IPV6][IN][BLOCK];
	total_in_block = ipv4_in_block + ipv6_in_block;

	total_in = total_in_pass + total_in_block;

	ipv4_out_pass = iface->pfif_bytes[IPV4][OUT][PASS];
	ipv6_out_pass = iface->pfif_bytes[IPV6][OUT][PASS];
	total_out_pass = ipv4_out_pass + ipv6_out_pass;

	ipv4_out_block = iface->pfif_bytes[IPV4][OUT][BLOCK];
	ipv6_out_block = iface->pfif_bytes[IPV6][OUT][BLOCK];
	total_out_block = ipv4_out_block + ipv6_out_block;

	total_out = total_out_pass + total_out_block;

	printf("%s In: Pass %llu Block %llu Total %llu\n", iface->pfif_name,
	    total_in_pass,
	    total_in_block,
	    total_in);

	printf("%s Out: Pass %llu Block %llu Total %llu\n", iface->pfif_name,
	    total_out_pass,
	    total_out_block,
	    total_out);
}

int get_pf_ifaces(int dev, const char *filter, struct pfi_if *buffer, int *numentries, int flags) {
	struct pfioc_iface io;

	bzero(&io, sizeof(io));	

	/* If a filter was specified, use it */
	if (filter != NULL && (strlcpy(io.pfiio_name, filter, sizeof(io.pfiio_name)) >= sizeof(io.pfiio_name))) {
		fprintf(stderr, "Interface string %s too long\n", filter);
		return (-1);
	}

	/* Set up our request structure */
	io.pfiio_buffer = buffer;	
	io.pfiio_esize = sizeof(*buffer);
	io.pfiio_size = *numentries;
	io.pfiio_flags = flags;
	
	if (ioctl(dev, DIOCIGETIFACES, &io) == -1) {
		warn("DIOCIGETIFACES failed:");
		return (-1);
	}

	/* Provide the number of entries to the caller */	
	*numentries = io.pfiio_size;
	
	return (0);
}

int main(int argc, char *argv[]) {
	struct pfi_if *ifaces = NULL;
	char *filter= NULL;
	int numentries = 0;
	int dev, i;

	if (argc == 2) {
		/* Use an iface filter, if provided */	
		filter = argv[1];
	} else if (argc != 1) {
		printf("Usage: %s [interface]\n", argv[0]);
		exit (1);
	}

	dev = open(pf_device, O_RDONLY);
	if (dev < 0) {
		err(1, "Opening %s failed", pf_device);
	}

	/* Get the number of pfi_if structs to be returned and malloc the required memory */
	if(get_pf_ifaces(dev, filter, ifaces, &numentries, PFI_FLAG_INSTANCE) < 0)
		errx(1, "get_pf_ifaces() failed");
		
	ifaces = xmalloc(sizeof(struct pfi_if) * numentries);

	/*
	 * Record the previous number of entries specified.
	 * If more are required after the second call, we'll have to realloc
	 * our ifaces buffer.
	 */
	i = numentries;
	while (1) {	
		if(get_pf_ifaces(dev, filter, ifaces, &numentries, PFI_FLAG_INSTANCE) < 0)
			errx(1, "get_pf_ifaces() failed");

		if (i < numentries) {
			/*
			 * An interface was added prior to this call, but after the last
			 * get_pf_ifaces call.
			 * Allocate more space and loop through again.
			 */
			i = numentries;
			xrealloc(ifaces, sizeof(struct pfi_if) * numentries);
		} else {
			/* The same number or fewer entries were returned */
			break;
		}
	}

	for(i = 0; i < numentries; i++)
		print_iface_stats(&ifaces[i]);

	close(dev);
	exit (0);
}
